"""Azure Device Models

This package provides object models for use within the Azure Provisioning Device SDK and Azure IoTHub Device SDK.
"""

from .x509 import X509  # noqa: F401
from .proxy_options import ProxyOptions  # noqa: F401
