# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""This module contains constants related to the pipeline package.
"""

# Feature names
C2D_MSG = "c2d"
INPUT_MSG = "input"
METHODS = "methods"
TWIN = "twin"
TWIN_PATCHES = "twin_patches"
